#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netlink.h>
#include <linux/random.h>
#include <net/sock.h>
#include "../../rtllib/rtllib.h"
#include "rtl_core.h"
#include "rtl_softap.h"
#include "rtl_p2p.h"
#ifdef P2P
static struct sock *p2p_nl_ksock = NULL;
static int g_pid = -1;
static struct net_device *net_dev;
void hex_dump(void *buf, int len)
{
    int i;

    for (i = 0 ; i < len; i++)
	 {
		 printk("%02x-", ((u8*)buf)[i]);
		 if (15 == i%16)
			 printk("\n");
	 }
    printk("\n");
}

/*------------------------------------------------------------------------------------------------
|					netlink mesg						|
--------------------------------------------------------------------------------------------------
|nlmsghdr1|	payload1	|nlmsghdr2|	payload2	|nlmsghdrx|	payloadx	|
--------------------------------------------------------------------------------------------------
|nlmsghdr |p2p_nl_skb_msg| data	|nlmsghdr |p2p_nl_skb_msg| data	|nlmsghdr |p2p_nl_skb_msg| data	|
--------------------------------------------------------------------------------------------------*/
int p2p_nl_send_msg(char *data_buf, u32 data_len, u8 event_id, u8 *sa)
{
	struct sk_buff *skb;
	struct nlmsghdr *nlh;
	struct drv_event *p_event = NULL;
	u32 len, skblen;
	int err;
	printk("====>%s(): event_id is %d\n",__func__,event_id);
	if (!p2p_nl_ksock || g_pid < 0) {
		err = -ENOENT;
		printk("======>Err!!! ENOENT\n");
		goto p2p_send_fail;
	}

	/*get skb_len*/
	len = P2P_NL_MSGALIGN(sizeof(struct drv_event) + data_len);
	skblen = NLMSG_SPACE(len);

	/*alloc skb*/
	skb = alloc_skb(skblen, GFP_ATOMIC);
	if (!skb) {
		err = -ENOBUFS;
		printk("======>Err!!!skb ENOBUFS\n");
		goto p2p_send_fail;
	}

	/*will call skb_put(skb, NLMSG_ALIGN(skblen - sizeof(*nlh)));*/
	/*nlmsg_put - Add a new netlink message to an skb*/
	/*put nlmsghdr nlh into skb, and get nlhtail=payload pointer*/
	nlh = nlmsg_put(skb, 0, 0, P2P_TRANSPORT_MSG,
				skblen - sizeof(*nlh), 0);
	if (!nlh) {
		err = -ENOBUFS;
		printk("======>Err!!!nlh ENOBUFS\n");
		kfree_skb(skb);
		goto p2p_send_fail;
	}
	p_event = NLMSG_DATA(nlh);
	memset(p_event, 0, sizeof(struct drv_event));
	p_event->event_id = event_id;
	if (sa) {
		memcpy(p_event->sa, sa, ETH_ALEN);
	}
	p_event->data_len = data_len;
	if (data_len > 0 && data_buf)
		memcpy(p_event->data, data_buf, data_len);
	
	err = nlmsg_unicast(p2p_nl_ksock, skb, g_pid);
	if (err) {
		/* nlmsg_unicast already kfree_skb'd */
		printk("======>nlmsg_unicast err\n");
		goto p2p_send_fail;
	}
	return 0;

p2p_send_fail:
	printk(KERN_WARNING
		"%s: Dropped P2P Msg : --err %d\n", __func__, err);
	return err;
}
static void p2p_send_pkt_from_uplayer(struct rtllib_device *ieee, struct p2p_nl_hdr *nl_hdr)
{
	struct rtllib_hdr_3addr *mac_hdr = NULL;
	struct sk_buff *new_skb = NULL;
	struct sta_info *pentry = NULL;
	struct cb_desc *tcb_desc = NULL;
	u8 *payload = (u8 *)(nl_hdr + 1);
	u8 oui[4] = {0x00, 0x50, 0xf2, 0x09};
	unsigned long flags, flags1;
	new_skb = dev_alloc_skb(ieee->tx_headroom + nl_hdr->msglen + sizeof( struct rtllib_hdr_3addr));
	if (new_skb == NULL) {
		printk("%s(): can't alloc skb\n",__FUNCTION__);
		return;
	}
	skb_reserve(new_skb, ieee->tx_headroom);
	tcb_desc = (cb_desc *)(new_skb->cb + 8);
	mac_hdr = (struct rtllib_hdr_3addr *)skb_put(new_skb, nl_hdr->msglen + sizeof( struct rtllib_hdr_3addr));
	mac_hdr->frame_ctl = cpu_to_le16(RTLLIB_STYPE_MANAGE_ACT);
	memcpy(mac_hdr->addr1, nl_hdr->addr1, ETH_ALEN);
	memcpy(mac_hdr->addr2, nl_hdr->addr2, ETH_ALEN);
	memcpy(mac_hdr->addr3, nl_hdr->addr3, ETH_ALEN);
	memcpy(mac_hdr->payload, payload,  nl_hdr->msglen);
	if ((payload[0] == 0x7f) && (memcmp(payload + 1, oui, 4) == 0) && payload[5] == 0x03) {
		printk("%s(): recv go disv req from utility\n",__func__);
		spin_lock_irqsave(&(ieee->sta_info_lock), flags);
		pentry = ap_get_stainfo(ieee, nl_hdr->addr1);
		if (pentry && pentry->bPowerSave) {
			tcb_desc->p2p_go_disv_req = true;
			spin_lock_irqsave(&(ieee->psQ_lock), flags1);
			skb_queue_tail(&(pentry->mgnt_ps_queue), new_skb);
			spin_unlock_irqrestore(&(ieee->psQ_lock), flags1);
			spin_unlock_irqrestore(&(ieee->sta_info_lock), flags);
			return;
		} else {
			spin_unlock_irqrestore(&(ieee->sta_info_lock), flags);
			p2p_nl_send_msg(NULL, 0, DRV_EVENT_GO_DISV_REQ_SENT, mac_hdr->addr1);
		}
	}
	if (ieee->p2p_role == P2P_ROLE_GO)
		softmac_mgmt_xmit(new_skb, ieee, true);
	else
		softmac_mgmt_xmit(new_skb, ieee, false);
	return;
}
static void p2p_nl_rcv_msg(struct sk_buff *skb)
{
	struct rtllib_device *ieee = (struct rtllib_device *)netdev_priv_rsl(net_dev);
	struct nlmsghdr *nlh;
	struct p2p_nl_hdr *hdr;
	u32 rlen;
	int err;

	while (skb->len >= NLMSG_SPACE(0)) {
		err = 0;

		/*get nlh*/
		nlh = nlmsg_hdr(skb);
		if ((nlh->nlmsg_len < (sizeof(*nlh) + sizeof(*hdr))) ||
		    (skb->len < nlh->nlmsg_len)) {
			printk(KERN_WARNING "%s: discarding partial skb\n",
				 __func__);
			return;
		}

		/*used to get a value that is larger than len and aligned*/
		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;

		if (nlh->nlmsg_type != P2P_TRANSPORT_MSG) {
			printk(KERN_WARNING "%s: nlmsg_type not P2P_TRANSPORT_MSG\n",
				 __func__);
			err = -EBADMSG;
			goto next_msg;
		}

		/*get p2p cmd header*/
		hdr = NLMSG_DATA(nlh);
		if (hdr->version != P2P_NL_VERSION || hdr->msgtype != P2P_NL_U2K) {
			printk(KERN_WARNING "%s: hdr wrong\n",
				 __func__);
			err = -EPROTOTYPE;
			goto next_msg;
		}
		g_pid = nlh->nlmsg_pid;

		if (nlh->nlmsg_len < (sizeof(*nlh) + hdr->msglen + sizeof(*hdr))) {
			printk(KERN_WARNING "%s: discarding partial message\n",
				 __func__);
			goto next_msg;
		}

		
		/*
		 * Deliver message to the appropriate transport
		 * free it temp
		 */
		p2p_send_pkt_from_uplayer(ieee, hdr);
next_msg:

		skb_pull(skb, rlen);
	}

	if(skb)
		;
}
u8 p2p_pending_client_entry_exist(struct rtllib_device *ieee, u8 *addr)
{
	u8 i = 0;
	for (i = 0; i < MAX_PENDING_CILENT_NUM; i++) {
		if (ieee->pending_client_list[i].used && (memcmp(ieee->pending_client_list[i].addr, addr, ETH_ALEN) == 0)) {
			break;
		}
	}
	return i;
}
u8 p2p_find_empty_pending_client_entry(struct rtllib_device *ieee)
{
	u8 i = 0;
	for (i = 0; i < MAX_PENDING_CILENT_NUM; i++) {
		if (ieee->pending_client_list[i].used == false) {
			return i;
		}
	}
	return MAX_PENDING_CILENT_NUM;
}
void p2p_del_pending_client_entry(struct rtllib_device *ieee, u8 idx)
{
	if (ieee->pending_client_list[idx].p2p_ie)
		kfree(ieee->pending_client_list[idx].p2p_ie);
	ieee->pending_client_list[idx].p2p_ie = NULL;
	memset((u8 *)&(ieee->pending_client_list[idx]), 0, sizeof(struct pending_client_info));
}
void p2p_pending_client_entry_age_timeout(struct rtllib_device *ieee)
{
	int                     idx = 0;
	unsigned long           CurrentTime;
	signed long             TimeDifference;
	CurrentTime = jiffies;

	for (idx = 0; idx < MAX_PENDING_CILENT_NUM; idx++) {
		if (ieee->pending_client_list[idx].used) {
			TimeDifference = jiffies_to_msecs(CurrentTime - ieee->pending_client_list[idx].start_time);
			if (TimeDifference > 60000) {
				printk("%s(): "MAC_FMT" timeout\n",__func__, MAC_ARG(ieee->pending_client_list[idx].addr));
				p2p_del_pending_client_entry(ieee, idx);
			}
		}
	}
}
void p2p_process_assoc_req(struct rtllib_device *ieee, struct sk_buff *skb)
{
	struct rtllib_assoc_request_frame * assoc_req= (struct rtllib_assoc_request_frame *)skb->data;
	struct rtllib_info_element *info_element = assoc_req->info_element;
	u16 length = skb->len - sizeof(struct rtllib_assoc_request_frame);
	int idx =0;
	printk("======>%s()\n",__func__);
	while (length >= sizeof(*info_element)){
		if (sizeof(*info_element) + info_element->len > length) {
			printk("Err! %s(): parse info_element err\n",__func__);
			break;
		}
		if (info_element->id == MFIE_TYPE_GENERIC && info_element->len > 4) {
			if(info_element->data[0] == 0x00 &&
				info_element->data[1] == 0x50 &&
				info_element->data[2] == 0xf2 &&
				info_element->data[3] == 0x09){
				printk("============>%s():has p2p IE\n",__FUNCTION__);
#if 1
				if (ieee->p2p_role == P2P_ROLE_GO && ieee->p2p_state == P2P_STATE_LINKED) {
					printk("@@@@@@@@@@ Send DRV-EVENT_NEW_CLIENT_JOIN\n");
					p2p_nl_send_msg(NULL, 0, DRV_EVENT_NEW_CLIENT_JOIN, assoc_req->header.addr2);
				}
#endif				
				idx = p2p_pending_client_entry_exist(ieee, assoc_req->header.addr2);
				printk("========>%s(): idx is %d\n",__func__,idx);
				if (idx < MAX_PENDING_CILENT_NUM) {
					printk("%s(): the pending client entry is already exist %d\n",__func__, idx);
					return;
				} else {
#if 0
					printk("p2p_role=%d state=%d\n", ieee->p2p_role, ieee->p2p_state);
					if (ieee->p2p_role == P2P_ROLE_GO && ieee->p2p_state == P2P_STATE_LINKED) {
						printk("=======> role is go and state is linked\n");
						if (ap_is_stainfo_exist(ieee, assoc_req->header.addr2) == APDEV_MAX_ASSOC) {
							printk("@@@@@@@@@@ Send DRV-EVENT_NEW_CLIENT_JOIN\n");
							p2p_nl_send_msg(NULL, 0, DRV_EVENT_NEW_CLIENT_JOIN, assoc_req->header.addr2);
						}
					}
#endif				
					idx = p2p_find_empty_pending_client_entry(ieee);
					printk("========>%s(): new idx is %d\n",__func__,idx);
					if (idx < MAX_PENDING_CILENT_NUM) {
						ieee->pending_client_list[idx].p2p_ie = kzalloc(info_element->len + 2, GFP_ATOMIC);
						if (ieee->pending_client_list[idx].p2p_ie == NULL) {
							printk("Err!%s(): can't alloc p2p_ie memory\n",__func__);
							return;
						}
						memcpy(ieee->pending_client_list[idx].p2p_ie, (u8 *)info_element, info_element->len + 2);
						ieee->pending_client_list[idx].p2p_ie_len = info_element->len + 2;
						memcpy(ieee->pending_client_list[idx].addr, assoc_req->header.addr2, ETH_ALEN);
						ieee->pending_client_list[idx].used = true;
						ieee->pending_client_list[idx].start_time = jiffies;
					} else
						printk("%s(): no more pending client list empty\n",__func__);
					return;	
				}
					
			}
		}
		length -= sizeof(struct rtllib_info_element) + info_element->len;
		info_element = (struct rtllib_info_element *)&(info_element->data[info_element->len]);
	}
}
bool p2p_parse_p2p_sub_ie_and_check_noa(u8 *data, u16 data_len, struct rtllib_network *pnetwork)
{
	u8 sub_ie_hdr = 2; 
	bool ret = false;
	u16 buf_len = data_len - 6; 
	u8 *buf = data + 6;
	while(buf_len >= sub_ie_hdr) {
		if (sub_ie_hdr + *(buf + 1) > buf_len) {
			printk("%s(): sub ie len(%d) larger than remain length(%d)\n",__func__,*(buf+1),buf_len);
			dump_buf(data, data_len);
			break;
		}
		switch (*buf) {
		case P2P_SUBIE_ID_CAPABILITY:
		{
			if (*(buf + 1) != 2) {
				RT_TRACE(COMP_ERR, "cap sub ie err\n");
				break;
			}
			pnetwork->cap_sub_ie_len = 4;
			memcpy(pnetwork->cap_sub_ie, buf, pnetwork->cap_sub_ie_len);
			break;
		}	
		case P2P_SUBIE_ID_DEVICE_ID:
		{
			if (*(buf + 1) != 6) {
				RT_TRACE(COMP_ERR, "dev id sub ie err\n");
				break;
			}
			pnetwork->dev_id_sub_ie_len = 8;
			memcpy(pnetwork->dev_id_sub_ie, buf, pnetwork->dev_id_sub_ie_len);
			break;
		}	
		case P2P_SUBIE_ID_NOA:
		{
			struct notice_of_absence_sub_ie *noa_ie = (struct notice_of_absence_sub_ie *)buf;
			u8 *p = noa_ie->noa;
			pnetwork->noa_info.index = noa_ie->index;
			pnetwork->noa_info.ctw_oppps = noa_ie->ctw_oppps;
			pnetwork->noa_info.noa.count_or_type = *p;
			pnetwork->noa_info.noa.duration = be32_to_cpu(*(u32 *)(p+1));
			pnetwork->noa_info.noa.interval = be32_to_cpu(*(u32 *)(p+5));
			pnetwork->noa_info.noa.start_time = be32_to_cpu(*(u32 *)(p+9));
			pnetwork->noa_info.noa.start_time = jiffies;
			printk("rx noa sub ie, start_time is %u, duration is %u\n",pnetwork->noa_info.noa.start_time,pnetwork->noa_info.noa.duration);
			pnetwork->noa_sub_ie_len = 17;
			memcpy(pnetwork->noa_sub_ie, buf, pnetwork->noa_sub_ie_len);
			ret = true;
			break;
		}
		case P2P_SUBIE_ID_EXT_LISTEN_TIMING:
		{
			if (*(buf + 1) != 4) {
				RT_TRACE(COMP_ERR, "ext listen sub ie err\n");
				break;
			}
			pnetwork->ext_listen_sub_ie_len = 6;
			memcpy(pnetwork->ext_listen_sub_ie, buf, pnetwork->ext_listen_sub_ie_len);
			break;
		}
		case P2P_SUBIE_ID_DEV_INFO:
		{
			pnetwork->dev_info_sub_ie_len = *(buf + 1) + 2;
			memcpy(pnetwork->dev_info_sub_ie, buf, pnetwork->dev_info_sub_ie_len);
			break;
		}
		case P2P_SUBIE_ID_GROUP_INFO:
		{
			pnetwork->group_info_sub_ie_len = *(buf + 1) + 2;
			memcpy(pnetwork->group_info_sub_ie, buf, pnetwork->group_info_sub_ie_len);
			break;
		}
		default:
			printk("%s() : unknown p2p sub ie(%d)\n",__func__,*buf);
			break;
		}
		buf_len -= sub_ie_hdr + *(buf + 1);
		buf = buf + sub_ie_hdr + *(buf + 1);
	}
	return ret;
}
u16 p2p_network_init(struct rtllib_device *ieee,
	struct p2p_network_t *p2p_net,
	struct rtllib_network *network)
{
	u8 *p = p2p_net->data;
	u8 p2p_ie_len = 0;
	if (network->p2p_ie_len > 0) {
		p2p_net->mode = NETWORK_MODE_P2P;
		if ((network->capability & 0x01) == 0)
			p2p_net->role = P2P_ROLE_DEVICE;
		else
			p2p_net->role = P2P_ROLE_GO;
	}
	p2p_net->channel = network->channel;
	memcpy(p2p_net->bssid, network->bssid, ETH_ALEN);
	memcpy(p2p_net->dev_addr, network->dev_addr, ETH_ALEN);
	memcpy(p2p_net->ssid, network->ssid, network->ssid_len);
	p2p_net->ssid_len = network->ssid_len;
	p2p_net->quality = network->SignalStrength;
	memcpy(p, (u8*)&network->wzc_ie_len, 1);
	p += 1;
	memcpy(p, network->wzc_ie, network->wzc_ie_len);
	p += network->wzc_ie_len;
	p2p_ie_len = network->cap_sub_ie_len + network->dev_id_sub_ie_len + network->noa_sub_ie_len
		+ network->ext_listen_sub_ie_len + network->dev_info_sub_ie_len + network->group_info_sub_ie_len;
	*p = p2p_ie_len + 6;
	p += 1;
	rtllib_build_p2p_ie_hdr(p, p2p_ie_len);
	p += 6;
	if (network->cap_sub_ie_len) {
		memcpy(p, network->cap_sub_ie, network->cap_sub_ie_len);
		p += network->cap_sub_ie_len;
	}
	if (network->dev_id_sub_ie_len) {
		memcpy(p, network->dev_id_sub_ie, network->dev_id_sub_ie_len);
		p += network->dev_id_sub_ie_len;
	}
	if (network->noa_sub_ie_len) {
		memcpy(p, network->noa_sub_ie, network->noa_sub_ie_len);
		p += network->noa_sub_ie_len;
	}
	if (network->ext_listen_sub_ie_len) {
		memcpy(p, network->ext_listen_sub_ie, network->ext_listen_sub_ie_len);
		p += network->ext_listen_sub_ie_len;
	}
	if (network->dev_info_sub_ie_len) {
		memcpy(p, network->dev_info_sub_ie, network->dev_info_sub_ie_len);
		p += network->dev_info_sub_ie_len;
	}
	if (network->group_info_sub_ie_len) {
		memcpy(p, network->group_info_sub_ie, network->group_info_sub_ie_len);
		p += network->group_info_sub_ie_len;
	}
	printk("mode=%d role=%d channel=%d ssid = %s",p2p_net->mode, p2p_net->role,p2p_net->channel, (char*)p2p_net->ssid);
	return (int)(p-(u8 *)p2p_net);
}
#define MAX_DISV_RESULT_LEN	1400
int p2p_indicate_disv_result(struct rtllib_device *ieee)
{
	struct rtllib_network *network;
	struct p2p_network_t p2p_net;
	u8 *buf = NULL;
	u16 buf_len = 0;
	int i = 0;
	u16 len = 0;
	bool only_one = true;
	bool bfirst = true;
	int ret = 0;
	printk("===>%s()\n",__func__);
	buf  = kzalloc(MAX_DISV_RESULT_LEN, GFP_ATOMIC);
	if (NULL == buf) {
		printk("%s(): can't malloc memory\n",__func__);
		ret = -ENOMEM;
		goto out;
	}
	memset(buf, 0, MAX_DISV_RESULT_LEN);
	list_for_each_entry(network, &ieee->network_list, list) {
		if (i == MAX_P2P_NETWORK_COUNT)
			break;
		if (ieee->scan_age == 0 ||
		    time_after(network->last_scanned + ieee->scan_age, jiffies)) {
			if (i == 0) {
				buf_len += 1;
			}
		 	memset((u8 *)&p2p_net, 0, sizeof(struct p2p_network_t));
			len = p2p_network_init(ieee, &p2p_net, network);
			if (buf_len + len + 2 > MAX_DISV_RESULT_LEN) {
				printk("=====>send_msg:buf_len is %d\n",buf_len);
				only_one = false;
				if (bfirst == true)
					buf[0] = 1;
				else
					buf[0] = 2;
				p2p_nl_send_msg(buf, buf_len, DRV_EVENT_SCAN_RESULTS_AVAILABLE, NULL);	
				bfirst = false;
				memset(buf, 0, MAX_DISV_RESULT_LEN);
				buf_len = 1;
			}
			memcpy(buf+buf_len, (u8 *)&len, 2);
			memcpy(buf+buf_len+2, (u8 *)&p2p_net, len);
			buf_len += len + 2;
			printk("=====>i=%d len is %d\n",i,len);
			i++;
		} else
			printk("Not showing network '%s ("
				MAC_FMT ")' due to age (%lums).\n",
				escape_essid(network->ssid,
					     network->ssid_len),
				MAC_ARG(network->bssid),
				(jiffies - network->last_scanned) / (HZ / 100));
		
	}
	if (buf_len > 0) {
		printk("=====>send_msg:buf_len is %d\n",buf_len);
		if (only_one == true)
			buf[0] = 0;
		else
			buf[0] = 3;
		p2p_nl_send_msg(buf, buf_len, DRV_EVENT_SCAN_RESULTS_AVAILABLE, NULL);	
	} else if (buf_len == 0 && i == 0) {
		printk("======>disv no networks\n");
		p2p_nl_send_msg(NULL, 0, DRV_EVENT_SCAN_RESULTS_AVAILABLE, NULL);	
	}

out:
	if (buf) {
		kfree(buf);
		buf = NULL;
	}
	return ret;
}
void p2p_discovery_timer_callback(unsigned long data)
{
	struct rtllib_device *ieee = (struct rtllib_device *) data;
	if (ieee->discover_hurryup && (ieee->disv_mode == P2P_DISV || ieee->disv_mode == P2P_GO_SCAN)){
		printk("============>discover_hurryup out\n");
		ieee->disv_mode = P2P_NONE;
		ieee->p2p_disv_phase = P2P_SCAN_PHASE;
		ieee->p2p_disv_channel = 0;
		ieee->find_circle_count = 0;
		ieee->discover_hurryup = false;
		if (ieee->disv_mode == P2P_GO_SCAN) {
		printk("======>%s():del noa timer\n",__func__);
			del_timer_sync(&ieee->p2p_noa_timer);
			cancel_delayed_work(&ieee->p2p_start_noa_wq);
			ieee->bin_noa = false;
			memset((u8 *)&ieee->noa, 0, sizeof(struct noa_desc));
			ieee->set_chan(ieee->dev, ieee->current_network.channel);
		}
		return;
	}
	if (ieee->p2p_dev_found && (ieee->disv_mode == P2P_SCAN ||
		ieee->disv_mode == P2P_SPEC_FIND || 
		ieee->disv_mode == P2P_ONLY_SEARCH)) {
		printk("============>dev found out\n");
		ieee->disv_mode = P2P_NONE;
		ieee->p2p_disv_phase = P2P_SCAN_PHASE;
		ieee->p2p_disv_channel = 0;
		ieee->find_circle_count = 0;
		ieee->discover_hurryup = false;
		ieee->p2p_dev_found = false;
		return;
	}
	down(&ieee->wx_sem);
	switch (ieee->p2p_disv_phase) {
	case P2P_SCAN_PHASE:
	{
		ieee->p2p_state = P2P_STATE_SCAN;
		down(&ieee->scan_sem);
		if (ieee->p2p_disv_channel + 1 > 11) {
			printk("%s(): scan complete\n",__func__);
			if (ieee->disv_mode == P2P_DISV) {
				ieee->p2p_disv_phase = P2P_LISTEN_PHASE;
				up(&ieee->scan_sem);
				up(&ieee->wx_sem);
				mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(1));
				return;
			} else {
				up(&ieee->scan_sem);
				up(&ieee->wx_sem);
				goto complete;
			}
			
		}
		ieee->p2p_disv_channel += 1;
		ieee->set_chan(ieee->dev, ieee->p2p_disv_channel);
		rtllib_send_probe_requests(ieee, 0);
		rtllib_send_probe_requests(ieee, 0);
		up(&ieee->scan_sem);
		mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(P2P_SCAN_TIME));
		break;
	}
	case P2P_LISTEN_PHASE:
	{
		u8 num = 0;
		ieee->p2p_state = P2P_STATE_LISTEN;
		get_random_bytes(&num, 1);
		num = (u8)num  % 3 + 1;
		ieee->set_chan(ieee->dev, ieee->listen_channel);
		ieee->p2p_disv_phase = P2P_SEARCH_PHASE;
		ieee->p2p_disv_channel = 0;
		mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(P2P_LISTEN_INTERVAL * num));
		break;
	}
	case P2P_SEARCH_PHASE:
	{
		ieee->p2p_state = P2P_STATE_SEARCH;
		switch (ieee->disv_mode) {
		case P2P_DISV:
		{
			if (ieee->p2p_disv_channel == 0)
				ieee->p2p_disv_channel = 1;
			else if (ieee->p2p_disv_channel == 1)
				ieee->p2p_disv_channel = 6;
			else if (ieee->p2p_disv_channel == 6)
				ieee->p2p_disv_channel = 11;
			else if (ieee->p2p_disv_channel == 11) {
				ieee->p2p_disv_channel = 0;
			 	if (ieee->find_circle_count == (P2P_FIND_PHASE_CIRCLE_COUNT - 1)) {
					ieee->find_circle_count = 0;
					printk("%s(): discovery complete\n",__func__);
					up(&ieee->wx_sem);
					goto complete;
				} else {
					ieee->p2p_disv_phase = P2P_LISTEN_PHASE;
					ieee->find_circle_count++;
					up(&ieee->wx_sem);
					mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(1));
					return;
				}
			
			}
			ieee->set_chan(ieee->dev, ieee->p2p_disv_channel);
			rtllib_send_probe_requests(ieee, 0);
			mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(P2P_SEARCH_TIME));
			break;
		}
		case P2P_SPEC_FIND:
		{
			if (ieee->find_circle_count == (P2P_FIND_PHASE_CIRCLE_COUNT - 1)) {
					ieee->find_circle_count = 0;
					printk("%s(): spec find complete\n",__func__);
					up(&ieee->wx_sem);
					goto complete;
			} else {
				ieee->p2p_disv_phase = P2P_LISTEN_PHASE;
				ieee->find_circle_count++;
				ieee->set_chan(ieee->dev, ieee->spec_channel);
				rtllib_send_probe_requests(ieee, 0);
				mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(P2P_SEARCH_TIME));
			}
			break;
		} 
		case P2P_ONLY_SEARCH:
		{
			if (ieee->p2p_disv_channel == 0)
				ieee->p2p_disv_channel = 1;
			else if (ieee->p2p_disv_channel == 1)
				ieee->p2p_disv_channel = 6;
			else if (ieee->p2p_disv_channel == 6)
				ieee->p2p_disv_channel = 11;
			else if (ieee->p2p_disv_channel == 11) {
				ieee->p2p_disv_channel = 0;
				printk("%s(): search complete\n",__func__);
				up(&ieee->wx_sem);
				goto complete;
				
			}
			ieee->set_chan(ieee->dev, ieee->p2p_disv_channel);
			rtllib_send_probe_requests(ieee, 0);
			rtllib_send_probe_requests(ieee, 0);
			mod_timer(&ieee->p2p_discovery_timer, jiffies + MSECS(P2P_SEARCH_TIME));
			break;
		}
		default:
			printk("%s(): known P2P DISV MODE %d\n",__func__, ieee->disv_mode);
			break;
		}
		break;
	}
	default:
		printk("%s(): known P2P PHASE %d\n",__func__, ieee->p2p_disv_phase);
		break;
	}
	up(&ieee->wx_sem);
	return;
complete:
	if (ieee->disv_mode == P2P_SPEC_FIND || ieee->disv_mode == P2P_SCAN) {
		printk("%s(): dev not found, disv_mode is %d\n",__func__,ieee->disv_mode);
		p2p_nl_send_msg(NULL, 0, DRV_EVENT_DEVICE_NOT_FOUND, ieee->assigned_addr);
       		if ((ieee->bp2p_mode) && (ieee->p2p_submode == 1) && ieee->need_backup_wlan) {
			HTSetConnectBwMode(ieee, (HT_CHANNEL_WIDTH)(ieee->client_chlwidth), (HT_EXTCHNL_OFFSET)(ieee->client_extchloffset), false);
			if(ieee->ScanOperationBackupHandler)
				ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_RESTORE);
			ieee->state = RTLLIB_LINKED;
			ieee->link_change(ieee->dev);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			ieee->need_backup_wlan = false;
		} 
	} else if (ieee->disv_mode == P2P_ONLY_SEARCH) {
		printk("%s(): dev not found, disv_mode is %d\n",__func__,ieee->disv_mode);
		p2p_nl_send_msg(NULL, 0, DRV_EVENT_DEVICE_NOT_FOUND, ieee->assigned_addr);
		ieee->bsearch = false;
		if (ieee->p2p_role == P2P_ROLE_CLIENT) {
			HTSetConnectBwMode(ieee, (HT_CHANNEL_WIDTH)(ieee->client_chlwidth), (HT_EXTCHNL_OFFSET)(ieee->client_extchloffset), false);
			if(ieee->ScanOperationBackupHandler)
				ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_RESTORE);
			ieee->state = RTLLIB_LINKED;
			ieee->link_change(ieee->dev);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			rtllib_sta_ps_send_null_frame(ieee, 0);
		}
		ieee->p2p_state = P2P_STATE_LINKED;
	} else if (ieee->disv_mode == P2P_DISV || ieee->disv_mode == P2P_GO_SCAN){
		if (ieee->p2p_role == P2P_ROLE_CLIENT) {
			HTSetConnectBwMode(ieee, (HT_CHANNEL_WIDTH)(ieee->client_chlwidth), (HT_EXTCHNL_OFFSET)(ieee->client_extchloffset), false);
			if(ieee->ScanOperationBackupHandler)
				ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_RESTORE);
			ieee->state = RTLLIB_LINKED;
			ieee->link_change(ieee->dev);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			ieee->p2p_state = P2P_STATE_LINKED;
		} 
		if ((ieee->bp2p_mode) && (ieee->p2p_submode == P2P_CONCURRENT) && ieee->need_backup_wlan) {
			HTSetConnectBwMode(ieee, (HT_CHANNEL_WIDTH)(ieee->client_chlwidth), (HT_EXTCHNL_OFFSET)(ieee->client_extchloffset), false);
			if(ieee->ScanOperationBackupHandler)
				ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_RESTORE);
			ieee->state = RTLLIB_LINKED;
			ieee->link_change(ieee->dev);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			rtllib_sta_ps_send_null_frame(ieee, 0);
			ieee->need_backup_wlan = false;
		} 
		if (ieee->disv_mode == P2P_GO_SCAN)
			ieee->p2p_state = P2P_STATE_LINKED;
		printk("===>%s(): end_time %u\n",__func__,jiffies_to_msecs(jiffies));
		p2p_indicate_disv_result(ieee);
		ieee->p2p_disv_phase = P2P_SCAN_PHASE;
		ieee->p2p_disv_channel = 0;
		ieee->find_circle_count = 0;
		ieee->discover_hurryup = false;
	}
	ieee->disv_mode = P2P_NONE;
	return;
	
}
void p2p_scan(struct rtllib_device *ieee)
{
	printk("===>%s()\n",__func__);
	del_timer_sync(&ieee->p2p_discovery_timer);
	ieee->wait_probe_rsp = false;
	ieee->p2p_dev_found = false;
	ieee->p2p_disv_phase = P2P_SCAN_PHASE;
	ieee->p2p_disv_channel=  0;
	ieee->find_circle_count = 0;
	ieee->disv_mode = P2P_SCAN;
	p2p_discovery_timer_callback((unsigned long)ieee);
}
void p2p_spec_find_phase(struct rtllib_device *ieee)
{
	printk("===>%s()\n",__func__);
	del_timer_sync(&ieee->p2p_discovery_timer);
	ieee->wait_probe_rsp = false;
	ieee->p2p_dev_found = false;
	ieee->p2p_disv_phase = P2P_LISTEN_PHASE;
	ieee->find_circle_count = 0;
	ieee->disv_mode = P2P_SPEC_FIND;
       	if ((ieee->bp2p_mode) && (ieee->p2p_submode == 1) && (ieee->state == RTLLIB_LINKED)) {
		ieee->need_backup_wlan = true;
		rtllib_sta_ps_send_null_frame(ieee, 1);
		rtllib_sta_ps_send_null_frame(ieee, 1);
		ieee->state = RTLLIB_LINKED_SCANNING;
        	ieee->link_change(ieee->dev);
		/* wait for ps packet to be kicked out successfully */
        	msleep(50);
		if(ieee->ScanOperationBackupHandler)
                	ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_BACKUP);
	}	
	p2p_discovery_timer_callback((unsigned long)ieee);
}
void p2p_discover(struct rtllib_device *ieee)
{
	printk("===>%s()\n",__func__);
	del_timer_sync(&ieee->p2p_discovery_timer);
	ieee->wait_probe_rsp = false;
	ieee->p2p_disv_phase = P2P_SCAN_PHASE;
	ieee->p2p_disv_channel = 0;
	ieee->find_circle_count = 0;
	ieee->disv_mode = P2P_DISV;
	if (ieee->p2p_role == P2P_ROLE_CLIENT && ieee->p2p_state == P2P_STATE_LINKED) {
		rtllib_sta_ps_send_null_frame(ieee, 1);
		rtllib_sta_ps_send_null_frame(ieee, 1);
		ieee->state = RTLLIB_LINKED_SCANNING;
        	ieee->link_change(ieee->dev);
		/* wait for ps packet to be kicked out successfully */
        	msleep(50);
		if(ieee->ScanOperationBackupHandler)
                	ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_BACKUP);
	}
       	if ((ieee->bp2p_mode) && (ieee->p2p_submode == 1) && (ieee->state == RTLLIB_LINKED)) {
		ieee->need_backup_wlan = true;
		rtllib_sta_ps_send_null_frame(ieee, 1);
		rtllib_sta_ps_send_null_frame(ieee, 1);
		ieee->state = RTLLIB_LINKED_SCANNING;
        	ieee->link_change(ieee->dev);
		/* wait for ps packet to be kicked out successfully */
        	msleep(50);
		if(ieee->ScanOperationBackupHandler)
                	ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_BACKUP);
	} 
	p2p_discovery_timer_callback((unsigned long)ieee);
}
void p2p_go_scan(struct rtllib_device *ieee)
{
	printk("===>%s()\n",__func__);
	del_timer_sync(&ieee->p2p_discovery_timer);
	ieee->wait_probe_rsp = false;
	ieee->p2p_disv_phase = P2P_SCAN_PHASE;
	ieee->p2p_disv_channel = 0;
       	if ((ieee->bp2p_mode) && (ieee->p2p_submode == 1) && (ieee->state == RTLLIB_LINKED)) {
		ieee->need_backup_wlan = true;
		rtllib_sta_ps_send_null_frame(ieee, 1);
		rtllib_sta_ps_send_null_frame(ieee, 1);
		ieee->state = RTLLIB_LINKED_SCANNING;
        	ieee->link_change(ieee->dev);
		/* wait for ps packet to be kicked out successfully */
        	msleep(50);
		if(ieee->ScanOperationBackupHandler)
                	ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_BACKUP);
	} 
	p2p_discovery_timer_callback((unsigned long)ieee);
}
void p2p_only_search(struct rtllib_device *ieee, bool is_go)
{
	printk("===>%s()\n",__func__);
	del_timer_sync(&ieee->p2p_discovery_timer);
	ieee->wait_probe_rsp = false;
	ieee->p2p_dev_found = false;
	ieee->p2p_disv_phase = P2P_SEARCH_PHASE;
	ieee->p2p_disv_channel = 0;
	ieee->disv_mode = P2P_ONLY_SEARCH;
	ieee->bsearch = true;
	if (!is_go) {
		rtllib_sta_ps_send_null_frame(ieee, 1);
		rtllib_sta_ps_send_null_frame(ieee, 1);
		ieee->state = RTLLIB_LINKED_SCANNING;
        	ieee->link_change(ieee->dev);
		/* wait for ps packet to be kicked out successfully */
        	msleep(50);
		if(ieee->ScanOperationBackupHandler)
                	ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_BACKUP);
	}
	p2p_discovery_timer_callback((unsigned long)ieee);
}
void p2p_deinit_variable(struct rtllib_device *ieee)
{
	del_timer_sync(&ieee->p2p_discovery_timer);
	printk("======>%s():del noa timer\n",__func__);
	del_timer_sync(&ieee->p2p_noa_timer);
	ieee->need_backup_wlan = false;
	ieee->wait_probe_rsp = false;
	ieee->ext_listen_hurryup = true;
	del_timer_sync(&ieee->p2p_ext_listen_timer);
	ieee->disv_mode = P2P_NONE;
	ieee->p2p_dev_found = false;
	ieee->p2p_disv_phase = 0;
	ieee->p2p_disv_channel = 0;
	ieee->bin_noa = false;
	cancel_delayed_work(&ieee->p2p_start_noa_wq);
	ieee->p2p_cap_ie_len = 0;
	memset(ieee->p2p_cap_ie, 0, P2P_SUBIE_MAX_LEN);
	ieee->dev_info_ie_len = 0;
	memset(ieee->dev_info_ie, 0, P2P_SUBIE_MAX_LEN);
	ieee->group_info_ie_len = 0;
	memset(ieee->group_info_ie, 0, P2P_SUBIE_MAX_LEN);
	ieee->dev_id_ie_len = 0;
	memset(ieee->dev_id_ie, 0, P2P_SUBIE_MAX_LEN);
	ieee->noa_ie_len = 0;
	memset(ieee->noa_ie, 0, P2P_SUBIE_MAX_LEN);
	ieee->ex_listen_ie_len = 0;
	memset(ieee->ex_listen_ie, 0, P2P_SUBIE_MAX_LEN);
	memset(ieee->wps_ie, 0, ieee->wps_ie_len);
	skb_queue_purge(&ieee->go_noa_queue);
}
int p2p_set_leave(struct rtllib_device *ieee, bool is_go)
{
	printk("===>%s()\n",__func__);
	rtllib_softmac_stop_protocol(ieee, 0, is_go, 1);
#if 0
	if (is_go) {
		apdev_down(ieee->apdev);
		_rtl8192_ap_up(ieee->dev, false);	
	}
#endif
	return 0;
}
void p2p_set_extend_listen_timer_callback(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct r8192_priv *priv = rtllib_priv(dev);
	struct rtllib_device* ieee = priv->rtllib;
	if (ieee->ext_listen_hurryup) {
		printk("=======>ieee->ext_listen_hurryup is true,return\n");
		ieee->ext_listen_hurryup = false;
		return ;
	}
	printk("===>%s()\n",__func__);
	if (ieee->p2p_state == P2P_STATE_IDLE) { 
		p2p_set_leave_idle(dev);
		ieee->set_chan(dev, ieee->listen_channel);
		ieee->p2p_state = P2P_STATE_LISTEN;
		mod_timer(&ieee->p2p_ext_listen_timer, jiffies + MSECS(P2P_EXT_LISTEN_PERIOD));
	} else if (ieee->p2p_state == P2P_STATE_LISTEN) {
		p2p_set_idle(dev);
		ieee->p2p_state = P2P_STATE_IDLE;
		mod_timer(&ieee->p2p_ext_listen_timer, jiffies + MSECS(P2P_EXT_LISTEN_INTERVAL-P2P_EXT_LISTEN_PERIOD));
	}
	return ;
}
void p2p_wait_invite_rsp_timer_callback(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct r8192_priv *priv = rtllib_priv(dev);
	struct rtllib_device* ieee = priv->rtllib;
	if (ieee->p2p_role != P2P_ROLE_CLIENT) {
		printk("%s(): p2p role is not client\n",__func__);
		return ;
	}
	printk("===>%s()\n",__func__);
	ieee->bsearch = false;
	HTSetConnectBwMode(ieee, (HT_CHANNEL_WIDTH)(ieee->client_chlwidth), (HT_EXTCHNL_OFFSET)(ieee->client_extchloffset), false);
	if(ieee->ScanOperationBackupHandler)
		ieee->ScanOperationBackupHandler(ieee->dev,SCAN_OPT_RESTORE);
	ieee->state = RTLLIB_LINKED;
	ieee->link_change(ieee->dev);
	rtllib_sta_ps_send_null_frame(ieee, 0);
	rtllib_sta_ps_send_null_frame(ieee, 0);
}
void p2p_noa_timer_callback(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct r8192_priv *priv = rtllib_priv(dev);
	struct rtllib_device* rtllib = priv->rtllib;
	printk("===>%s(): time is %u\n",__func__,jiffies_to_msecs(jiffies));
	rtllib->bin_noa = false;
	memset((u8 *)&rtllib->noa, 0, sizeof(struct noa_desc));
	if (rtllib->iw_mode == IW_MODE_MASTER || 
			(rtllib->iw_mode == IW_MODE_APSTA && rtllib->state == RTLLIB_NOLINK)) {
		if(rtllib->pAPHTInfo->bRegBW40MHz)
			HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20_40, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
		else
			HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
		rtllib->APExtChlOffset = (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER;
	} else {
		if ((rtllib->mode & (WIRELESS_MODE_N_24G |WIRELESS_MODE_N_5G)) == 0) {
			printk("===>sta is not N mode , set BW which is AP supported\n");
			rtllib->APExtChlOffset = (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER;
			if(rtllib->pAPHTInfo->bRegBW40MHz)
				HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20_40, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
			else
				HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
		} else {
			printk("===>sta is N mode\n");
			if (rtllib->pHTInfo->bCurBW40MHz == false) {
				printk("==========>sta is 20M\n");
				rtllib->APExtChlOffset = (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER;
				if(rtllib->pAPHTInfo->bRegBW40MHz)
					HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20_40, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
				else
					HTSetConnectBwMode(rtllib, HT_CHANNEL_WIDTH_20, (rtllib->current_ap_network.channel<=6)?HT_EXTCHNL_OFFSET_UPPER:HT_EXTCHNL_OFFSET_LOWER, true);  
			} else
				HTSetConnectBwMode(rtllib, (HT_CHANNEL_WIDTH)(rtllib->client_chlwidth), (HT_EXTCHNL_OFFSET)(rtllib->client_extchloffset), false);
		}
	}
}
int p2p_set_idle(struct net_device *dev)
{
	struct r8192_priv *priv = rtllib_priv(dev);
	unsigned long flags = 0;
#ifdef CONFIG_ASPM_OR_D3
	PRT_POWER_SAVE_CONTROL	pPSC = (PRT_POWER_SAVE_CONTROL)(&(priv->rtllib->PowerSaveControl));
#endif	
	printk("===>%s()\n",__func__);
	if (priv->rtllib->p2p_submode == P2P_CONCURRENT) {
		rtllib_softmac_stop_protocol(priv->rtllib, 0, 0, 1);	
	}
	spin_lock_irqsave(&priv->rf_ps_lock,flags);
	if (priv->RFChangeInProgress) {
		spin_unlock_irqrestore(&priv->rf_ps_lock,flags);
		RT_TRACE(COMP_RF, "rtl8192_hw_sleep_down(): RF Change in progress! \n");
		printk("rtl8192_hw_sleep_down(): RF Change in progress!\n");
		return -1;
	}
	spin_unlock_irqrestore(&priv->rf_ps_lock,flags);
	RT_TRACE(COMP_PS, "%s()============>come to sleep down\n", __FUNCTION__);

	MgntActSet_RF_State(dev, eRfSleep, RF_CHANGE_BY_IDLE,false);
#ifdef CONFIG_ASPM_OR_D3
	if(pPSC->RegRfPsLevel & RT_RF_OFF_LEVL_ASPM)
	{
		RT_ENABLE_ASPM(dev);
		RT_SET_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_ASPM);
	}
#ifdef TODO		
	else if(pPSC->RegRfPsLevel & RT_RF_OFF_LEVL_PCI_D3)
	{
		RT_ENTER_D3(dev, false);
		RT_SET_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_PCI_D3);
	}
#endif		
#endif
	priv->rtllib->p2p_state = P2P_STATE_IDLE;
	return 0;
}
int p2p_set_leave_idle(struct net_device *dev)
{
	struct r8192_priv *priv = rtllib_priv(dev);
	unsigned long flags = 0;
#ifdef CONFIG_ASPM_OR_D3
	PRT_POWER_SAVE_CONTROL	pPSC = (PRT_POWER_SAVE_CONTROL)(&(priv->rtllib->PowerSaveControl));
#endif	
	printk("===>%s()\n",__func__);
	spin_lock_irqsave(&priv->rf_ps_lock,flags);
	if (priv->RFChangeInProgress) {
		spin_unlock_irqrestore(&priv->rf_ps_lock,flags);
		RT_TRACE(COMP_RF, "rtl8192_hw_wakeup(): RF Change in progress! \n");
		printk("rtl8192_hw_wakeup(): RF Change in progress! schedule wake up task again\n");
		queue_delayed_work_rsl(priv->rtllib->wq,&priv->rtllib->hw_wakeup_wq,MSECS(10));
		return -1;
	}
	spin_unlock_irqrestore(&priv->rf_ps_lock,flags);
#ifdef CONFIG_ASPM_OR_D3
	if((pPSC->RegRfPsLevel & RT_RF_OFF_LEVL_ASPM) && RT_IN_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_ASPM))
	{
		RT_DISABLE_ASPM(dev);
		RT_CLEAR_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_ASPM);
	}
#ifdef TODO		
	else if((pPSC->RegRfPsLevel & RT_RF_OFF_LEVL_PCI_D3) && RT_IN_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_PCI_D3))
	{
		RT_LEAVE_D3(dev, false);
		RT_CLEAR_PS_LEVEL(pPSC, RT_RF_OFF_LEVL_PCI_D3);
	}
#endif		
#endif
	RT_TRACE(COMP_PS, "%s()============>come to wake up\n", __FUNCTION__);
	MgntActSet_RF_State(dev, eRfOn, RF_CHANGE_BY_IDLE,false);
	return 0;
}
void p2p_start_noa(void *data)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
	struct rtllib_device *ieee = container_of_work_rsl(data,struct rtllib_device, p2p_start_noa_wq);
#else
	struct net_device *dev = (struct net_device *)data;
#endif
	printk("===>%s(): time is %lu\n",__func__, MSECS(jiffies));
	ieee->bin_noa = true;
	if (ieee->disv_mode == P2P_GO_SCAN)
		p2p_go_scan(ieee);
	else if (ieee->disv_mode == P2P_ONLY_SEARCH) {
		if (ieee->p2p_role == P2P_ROLE_GO)
			p2p_only_search(ieee, true);
	}
}
int p2p_nl_init(struct net_device *dev)
{
	net_dev = dev;
	p2p_nl_ksock = netlink_kernel_create(&init_net, 
		P2P_NETLINK_TYPE, P2P_NL_GRP,
		p2p_nl_rcv_msg, NULL, 
		THIS_MODULE);

	if(p2p_nl_ksock) {
		printk("%s: alloc nl_sock ok\n", __func__);
		return 0;
	} else {
		printk("%s: alloc nl_sock err\n", __func__);
		return -ENOMEM; 
	}
}
void p2p_nl_exit(void)
{
	printk("%s: start\n", __func__);
	netlink_kernel_release(p2p_nl_ksock); 
	printk("%s: endi\n", __func__);
}
#endif
